import re
import struct
from django import forms
from django.core.validators import RegexValidator
from django.db import models, connection
from django.utils.translation import ugettext_lazy as _

try:
	from django.utils.six import with_metaclass
except ImportError:
	from six import with_metaclass


__all__ = ["HexadecimalField", "HexIntegerField"]

hex_re = re.compile(r"^0x[0-9a-fA-F]+$")
postgres_engines = [
	"django.db.backends.postgresql_psycopg2",
	"django.contrib.gis.db.backends.postgis",
]


class HexadecimalField(forms.CharField):
	"""
	A form field that accepts only hexadecimal numbers
	"""
	def __init__(self, *args, **kwargs):
		self.default_validators = [RegexValidator(hex_re, _("Enter a valid hexadecimal number"), "invalid")]
		super(HexadecimalField, self).__init__(*args, **kwargs)


class HexIntegerField(with_metaclass(models.SubfieldBase, models.BigIntegerField)):
	"""
	This field stores a hexadecimal *string* of up to 64 bits as an unsigned integer
	on *all* backends including postgres.

	Reasoning: Postgres only supports signed bigints. Since we don't care about
	signedness, we store it as signed, and cast it to unsigned when we deal with
	the actual value (with struct)

	On sqlite and mysql, native unsigned bigint types are used. In all cases, the
	value we deal with in python is always in hex.
	"""
	def db_type(self, connection):
		engine = connection.settings_dict["ENGINE"]
		if engine == "django.db.backends.mysql":
			return "bigint unsigned"
		elif engine == "django.db.backends.sqlite":
			return "UNSIGNED BIG INT"
		else:
			return super(HexIntegerField, self).db_type(connection)

	def get_prep_value(self, value):
		if value is None or value == "":
			return None
		value = int(value, 16)
		# on postgres only, interpret as signed
		if connection.settings_dict["ENGINE"] in postgres_engines:
			value = struct.unpack("q", struct.pack("Q", value))[0]
		return value

	def to_python(self, value):
		if isinstance(value, str):
			return value
		if value is None:
			return ""
		# on postgres only, re-interpret from signed to unsigned
		if connection.settings_dict["ENGINE"] in postgres_engines:
			value = hex(struct.unpack("Q", struct.pack("q", value))[0])
		return value

	def formfield(self, **kwargs):
		defaults = {"form_class": HexadecimalField}
		defaults.update(kwargs)
		# yes, that super call is right
		return super(models.IntegerField, self).formfield(**defaults)

try:
	from south.modelsinspector import add_introspection_rules
	add_introspection_rules([], ["^push_notifications\.fields\.HexIntegerField"])
except ImportError:
	pass



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from .models import APNSDevice, GCMDevice


class DeviceAdmin(admin.ModelAdmin):
	list_display = ("__unicode__", "device_id", "user", "active", "date_created")
	search_fields = ("name", "device_id", "user__username")
	list_filter = ("active", )
	actions = ("send_message", "send_bulk_message", "enable", "disable")

	def send_message(self, request, queryset):
		ret = []
		errors = []
		r = ""
		for device in queryset:
			try:
				r = device.send_message("Test single notification")
			except Exception as e:
				errors.append(str(e))
			if r:
				ret.append(r)
		if errors:
			self.message_user(request, _("Some messages could not be processed: %r" % ("\n".join(errors))))
		if ret:
			self.message_user(request, _("All messages were sent: %s" % ("\n".join(ret))))
	send_message.short_description = _("Send test message")

	def send_bulk_message(self, request, queryset):
		r = queryset.send_message("Test bulk notification")
		self.message_user(request, _("All messages were sent: %s" % (r)))
	send_bulk_message.short_description = _("Send test message in bulk")

	def enable(self, request, queryset):
		queryset.update(active=True)
	enable.short_description = _("Enable selected devices")

	def disable(self, request, queryset):
		queryset.update(active=False)
	disable.short_description = _("Disable selected devices")


admin.site.register(APNSDevice, DeviceAdmin)
admin.site.register(GCMDevice, DeviceAdmin)



"""
Google Cloud Messaging
Previously known as C2DM
Documentation is available on the Android Developer website:
https://developer.android.com/google/gcm/index.html
"""

import json

try:
	from urllib.request import Request, urlopen
	from urllib.parse import urlencode
except ImportError:
	# Python 2 support
	from urllib2 import Request, urlopen
	from urllib import urlencode

from django.core.exceptions import ImproperlyConfigured
from . import NotificationError
from .settings import PUSH_NOTIFICATIONS_SETTINGS as SETTINGS


class GCMError(NotificationError):
	pass


def _chunks(l, n):
	"""
	Yield successive chunks from list \a l with a minimum size \a n
	"""
	for i in range(0, len(l), n):
		yield l[i:i + n]


def _gcm_send(data, content_type):
	key = SETTINGS.get("GCM_API_KEY")
	if not key:
		raise ImproperlyConfigured('You need to set PUSH_NOTIFICATIONS_SETTINGS["GCM_API_KEY"] to send messages through GCM.')

	headers = {
		"Content-Type": content_type,
		"Authorization": "key=%s" % (key),
		"Content-Length": str(len(data)),
	}

	request = Request(SETTINGS["GCM_POST_URL"], data, headers)
	response = urlopen(request)
	result = response.read().decode("utf-8")

	# FIXME: broken for bulk results
	if result.startswith("Error="):
		raise GCMError(result)

	return result


def gcm_send_message(registration_id, data, collapse_key=None, delay_while_idle=False):
	"""
	Sends a GCM notification to a single registration_id.
	This will send the notification as form data.
	If sending multiple notifications, it is more efficient to use
	gcm_send_bulk_message() with a list of registration_ids
	"""

	values = {"registration_id": registration_id}

	if collapse_key:
		values["collapse_key"] = collapse_key

	for k, v in data.items():
		values["data.%s" % (k)] = v.encode("utf-8")

	data = urlencode(values).encode("utf-8")
	return _gcm_send(data, "application/x-www-form-urlencoded;charset=UTF-8")


def gcm_send_bulk_message(registration_ids, data, collapse_key=None, delay_while_idle=False):
	"""
	Sends a GCM notification to one or more registration_ids. The registration_ids
	needs to be a list.
	This will send the notification as json data.
	"""

	# GCM only allows up to 1000 reg ids per bulk message
	# https://developer.android.com/google/gcm/gcm.html#request
	max_recipients = SETTINGS.get("GCM_MAX_RECIPIENTS")
	if len(registration_ids) > max_recipients:
		ret = []
		for chunk in _chunks(registration_ids, max_recipients):
			ret.append(gcm_send_bulk_message(chunk, data, collapse_key, delay_while_idle))
		return "\n".join(ret)

	values = {"registration_ids": registration_ids}

	if data is not None:
		values["data"] = data

	if collapse_key:
		values["collapse_key"] = collapse_key,

	if delay_while_idle:
		values["delay_while_idle"] = delay_while_idle

	data = json.dumps(values, separators=(",", ":")).encode("utf-8")
	return _gcm_send(data, "application/json")



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration

from django.db import models
from django.conf import settings

try:
	from django.contrib.auth import get_user_model
except ImportError:  # django < 1.5
	from django.contrib.auth.models import User
else:
	User = get_user_model()

AUTH_USER_MODEL = getattr(settings, "AUTH_USER_MODEL", "auth.User")


class Migration(SchemaMigration):
	def forwards(self, orm):
		# Adding model "GCMDevice"
		db.create_table(u"push_notifications_gcmdevice", (
			("id", self.gf("django.db.models.fields.AutoField")(primary_key=True)),
			("name", self.gf("django.db.models.fields.CharField")(max_length=255, null=True, blank=True)),
			("active", self.gf("django.db.models.fields.BooleanField")(default=True)),
			("user", self.gf("django.db.models.fields.related.ForeignKey")(
				to=orm["%s.%s" % (User._meta.app_label, User._meta.object_name)], null=True, blank=True)),
			("device_id", self.gf("push_notifications.fields.HexIntegerField")(null=True, blank=True)),
			("registration_id", self.gf("django.db.models.fields.TextField")()),
		))
		db.send_create_signal(u"push_notifications", ["GCMDevice"])

		# Adding model "APNSDevice"
		db.create_table(u"push_notifications_apnsdevice", (
			("id", self.gf("django.db.models.fields.AutoField")(primary_key=True)),
			("name", self.gf("django.db.models.fields.CharField")(max_length=255, null=True, blank=True)),
			("active", self.gf("django.db.models.fields.BooleanField")(default=True)),
			("user", self.gf("django.db.models.fields.related.ForeignKey")(
				to=orm["%s.%s" % (User._meta.app_label, User._meta.object_name)], null=True, blank=True)),
			("device_id", self.gf("uuidfield.fields.UUIDField")(max_length=32, null=True, blank=True)),
			("registration_id", self.gf("django.db.models.fields.CharField")(unique=True, max_length=64)),
		))
		db.send_create_signal(u"push_notifications", ["APNSDevice"])

	def backwards(self, orm):
		# Deleting model "GCMDevice"
		db.delete_table(u"push_notifications_gcmdevice")

		# Deleting model "APNSDevice"
		db.delete_table(u"push_notifications_apnsdevice")

	models = {
		u"auth.group": {
			"Meta": {"object_name": "Group"},
			"id": ("django.db.models.fields.AutoField", [], {"primary_key": "True"}),
			"name": ("django.db.models.fields.CharField", [], {"unique": "True", "max_length": "80"}),
			"permissions": ("django.db.models.fields.related.ManyToManyField", [],
				{"to": u"orm['auth.Permission']", "symmetrical": "False", "blank": "True"})
		},
		u"auth.permission": {
			"Meta": {"ordering": "(u'content_type__app_label', u'content_type__model', u'codename')",
				"unique_together": "((u'content_type', u'codename'),)", "object_name": "Permission"},
			"codename": ("django.db.models.fields.CharField", [], {"max_length": "100"}),
			"content_type": ("django.db.models.fields.related.ForeignKey", [], {"to": u"orm['contenttypes.ContentType']"}),
			"id": ("django.db.models.fields.AutoField", [], {"primary_key": "True"}),
			"name": ("django.db.models.fields.CharField", [], {"max_length": "50"})
		},
		"%s.%s" % (User._meta.app_label, User._meta.module_name): {
			"Meta": {"object_name": User.__name__, 'db_table': "'%s'" % User._meta.db_table},
			"date_joined": ("django.db.models.fields.DateTimeField", [], {"default": "datetime.datetime.now"}),
			"email": ("django.db.models.fields.EmailField", [], {"max_length": "75", "blank": "True"}),
			"first_name": ("django.db.models.fields.CharField", [], {"max_length": "30", "blank": "True"}),
			"groups": ("django.db.models.fields.related.ManyToManyField", [],
				{"to": u"orm['auth.Group']", "symmetrical": "False", "blank": "True"}),
			"id": ("django.db.models.fields.AutoField", [], {"primary_key": "True"}),
			"is_active": ("django.db.models.fields.BooleanField", [], {"default": "True"}),
			"is_staff": ("django.db.models.fields.BooleanField", [], {"default": "False"}),
			"is_superuser": ("django.db.models.fields.BooleanField", [], {"default": "False"}),
			"last_login": ("django.db.models.fields.DateTimeField", [], {"default": "datetime.datetime.now"}),
			"last_name": ("django.db.models.fields.CharField", [], {"max_length": "30", "blank": "True"}),
			"password": ("django.db.models.fields.CharField", [], {"max_length": "128"}),
			"user_permissions": ("django.db.models.fields.related.ManyToManyField", [],
				{"to": u"orm['auth.Permission']", "symmetrical": "False", "blank": "True"}),
			"username": ("django.db.models.fields.CharField", [], {"unique": "True", "max_length": "30"})
		},
		u"contenttypes.contenttype": {
			"Meta": {"ordering": "('name',)", "unique_together": "(('app_label', 'model'),)", "object_name": "ContentType",
				"db_table": "'django_content_type'"},
			"app_label": ("django.db.models.fields.CharField", [], {"max_length": "100"}),
			"id": ("django.db.models.fields.AutoField", [], {"primary_key": "True"}),
			"model": ("django.db.models.fields.CharField", [], {"max_length": "100"}),
			"name": ("django.db.models.fields.CharField", [], {"max_length": "100"})
		},
		u"push_notifications.apnsdevice": {
			"Meta": {"object_name": "APNSDevice"},
			"active": ("django.db.models.fields.BooleanField", [], {"default": "True"}),
			"device_id": ("uuidfield.fields.UUIDField", [], {"max_length": "32", "null": "True", "blank": "True"}),
			"id": ("django.db.models.fields.AutoField", [], {"primary_key": "True"}),
			"name": ("django.db.models.fields.CharField", [], {"max_length": "255", "null": "True", "blank": "True"}),
			"registration_id": ("django.db.models.fields.CharField", [], {"unique": "True", "max_length": "64"}),
			"user": ("django.db.models.fields.related.ForeignKey", [],
				{"to": u"orm['%s.%s']" % (User._meta.app_label, User._meta.object_name), "null": "True", "blank": "True"})
		},
		u"push_notifications.gcmdevice": {
			"Meta": {"object_name": "GCMDevice"},
			"active": ("django.db.models.fields.BooleanField", [], {"default": "True"}),
			"device_id": ("push_notifications.fields.HexIntegerField", [], {"null": "True", "blank": "True"}),
			"id": ("django.db.models.fields.AutoField", [], {"primary_key": "True"}),
			"name": ("django.db.models.fields.CharField", [], {"max_length": "255", "null": "True", "blank": "True"}),
			"registration_id": ("django.db.models.fields.TextField", [], {}),
			"user": ("django.db.models.fields.related.ForeignKey", [],
				{"to": u"orm['%s.%s']" % (User._meta.app_label, User._meta.object_name), "null": "True", "blank": "True"})
		}
	}

	complete_apps = ["push_notifications"]
